import random
import time
import threading
from graphics_and_games_klassen import Rechteck, Text, Dreieck, Ereignisbehandlung

class PfeilSymbol:
    """
    Stellt einen Pfeil für den Vorratsstapel dar.
    """
    def __init__(self, hoehe, tiefe):
        """
        Legt das Pfeilsymbol an.
        :param hoehe: Anzeigehöhe (Pfeilspitze)
        :param tiefe: die anzuzeigende Rekursionstiefe
        """
        self.linie = Rechteck()
        self.linie.PositionSetzen(210, hoehe)
        self.linie.GroesseSetzen(30, 1)
        self.linie.FarbeSetzen("schwarz")
        self.spitze = Dreieck()
        self.spitze.GroesseSetzen(10, 10)
        self.spitze.PositionSetzen(210, hoehe - 5)
        self.spitze.FarbeSetzen("schwarz")
        self.spitze.Drehen(90)
        self.beschriftung = Text()
        self.beschriftung.PositionSetzen(245, hoehe + 4)
        self.beschriftung.TextSetzen(str(tiefe))

    def Entfernen(self):
        """
        Entfernt das Pfeilsymbol aus der Anzeige
        """
        self.linie.Entfernen()
        self.spitze.Entfernen()
        self.beschriftung.Entfernen()

    def HoeheSetzen(self, hoehe):
        """
        Setzt das Symbol auf eine neue Höhe
        :param hoehe: die neue Höhe
        """
        self.linie.PositionSetzen(210, hoehe)
        self.spitze.PositionSetzen(210, hoehe - 5)
        self.beschriftung.PositionSetzen(245, hoehe + 4)

class PalettenSymbol:
    """
    Stellt eine Palette auf der Zeichenfläche dar.
    """
    def __init__(self, x, y, hoehe, gewicht, farbe):
        """
        Legt die Elemente des Symbols an
        :param x: x-Koordinate
        :param y: y-Koordinate
        :param hoehe: Höhe der Palette
        :param gewicht: Gewicht für die Beschriftung
        :param farbe: Farbe
        """
        self.x = x
        self.y = y
        self.hoehe = hoehe
        self.farbe = farbe
        self.gewicht = gewicht
        self.rahmen = Rechteck()
        self.rahmen.PositionSetzen(x, y)
        self.rahmen.GroesseSetzen(100, hoehe)
        self.rahmen.FarbeSetzen("schwarz")
        self.fläche = Rechteck()
        self.fläche.PositionSetzen(x + 2, y + 2)
        self.fläche.GroesseSetzen(96, hoehe - 4)
        self.fläche.FarbeSetzen(farbe)
        self.text = Text()
        self.text.PositionSetzen(x + 40, y + hoehe - 4)
        self.text.TextSetzen(str(gewicht))

    def KopieErstellen(self):
        """
        Erzeugt ein neues Symbol mit gleichen Daten
        :return: Kopie
        """
        return PalettenSymbol(self.x, self.y, self.hoehe, self.gewicht, self.farbe)

    def PositionSetzen(self, x, y):
        """
        Setzt die Position (der linken oberen Ecke) des Symbols.
        :param x: x-Position der linken oberen Ecke
        :param y: y-Position der linken oberen Ecke
        """
        self.x = x
        self.y = y
        self.rahmen.PositionSetzen(x, y)
        self.fläche.PositionSetzen(x + 2, y + 2)
        self.text.PositionSetzen(x + 40, y + self.hoehe - 4)

    def XPositionSetzen(self, x):
        """
        Setzt die Position (der linken oberen Ecke) des Symbols.
        :param x: x-Position der linken oberen Ecke
        """
        self.PositionSetzen(x, self.y)

    def HoeheGeben(self):
        """
        Gibt die Höhe des Symbols zurück.
        :return: Höhe
        """
        return self.hoehe

    def Entfernen(self):
        """
        Entfernt das Symbol aus der Anzeige
        """
        self.rahmen.Entfernen()
        self.fläche.Entfernen()
        self.text.Entfernen()

class Palette:
    """
    Beschreibt die Daten einer Palette
    """
    def __init__(self, h, g):
        """
        Der Konstruktor legt die Daten fest
        :param h: Höhe der Palette
        :param g: Gewicht der Palette
        """
        self.hoehe = h
        self.gewicht = g

    def HoeheGeben(self):
        """
        Gibt die Höhe der Palette zurück
        :return: Höhe
        """
        return self.hoehe

    def GewichtGeben(self):
        """
        Gibt das Gewicht der Palette zurück
        :return: Gewicht
        """
        return self.gewicht
        

class Stapelverwaltung(Ereignisbehandlung):
    """
    Verwaltet die Anzeige der Palettenstapel
    """
    def __init__(self, vorrat, maximaleHoehe):
        """
        Legt den Vorratsstapel an
        :param vorrat: der Palettenvorrat
        :param maximaleHoehe: die maximale Höhe, die ein Palettenstapel haben darf
        """
        super().__init__()
        self.farben = ["rot", "grün", "blau", "gelb", "cyan", "magenta", "orange", "braun", "weiß"]
        self.vorratsStapel = [PalettenSymbol(100, 450 - sum(p.HoeheGeben() for p in vorrat[:i+1]), p.HoeheGeben(), p.GewichtGeben(), self.farben[i % len(self.farben)]) for i, p in enumerate(vorrat)]
        self.aktuellerStapel = []
        self.optimalerStapel = []
        self.pfeile = []
        self.tickDa = False
        self.bezeichnerVorrat = Text()
        self.bezeichnerVorrat.TextSetzen("Palettenvorrat")
        self.bezeichnerVorrat.PositionSetzen(110, 470)
        self.bezeichnerAktuell = Text()
        self.bezeichnerAktuell.TextSetzen("Aktueller Versuch")
        self.bezeichnerAktuell.PositionSetzen(400, 470)
        self.bezeichnerOptimal = Text()
        self.bezeichnerOptimal.TextSetzen("Bisher bester Stapel")
        self.bezeichnerOptimal.PositionSetzen(600, 470)
        self.oberkante = Rechteck()
        self.oberkante.GroesseSetzen(400, 3)
        self.oberkante.PositionSetzen(400, 450 - maximaleHoehe - 3)
        self.oberkante.FarbeSetzen("schwarz")
        self.bezeichnerOberkante = Text()
        self.bezeichnerOberkante.TextSetzen("Maximale Stapelhöhe")
        self.bezeichnerOberkante.PositionSetzen(400, 450 - maximaleHoehe - 20)
        

    def AufTaktimpulsWarten(self):
        """
        Wartet auf den nächsten Tick des Taktgebers
        """
        time.sleep(1)
        


    def AktionAusfuehren(self):
        """
        Die Aktionsmethode des Zeitgebers
        """
        self.tickDa = True

    def OptimalenStapelSetzen(self):
        """
        Setzt den optimalen Stapel
        pre: der aktuelle Stapel ist optimal.
        """
        for s in self.optimalerStapel:
            s.Entfernen()
        self.optimalerStapel.clear()
        for s in self.aktuellerStapel:
            neu = s.KopieErstellen()
            neu.XPositionSetzen(600)
            self.optimalerStapel.append(neu)
        self.AufTaktimpulsWarten()

    def AktuellHinzufügen(self, index):
        """
        Fügt den Eintrag für die Palette mit dem gegebenen Index zum aktuellen Stapel hinzu
        :param index: der Index der Palette
        """
        neu = self.vorratsStapel[index].KopieErstellen()
        h = sum(s.HoeheGeben() for s in self.aktuellerStapel)
        neu.PositionSetzen(400, 450 - (h + neu.HoeheGeben()))
        self.aktuellerStapel.append(neu)
        self.AufTaktimpulsWarten()

    def AktuellEntfernen(self):
        """
        Entfernt das oberste Element vom aktuellen Stapel.
        """
        alt = self.aktuellerStapel.pop()
        alt.Entfernen()
        self.AufTaktimpulsWarten()

    def PfeilSetzen(self, index, tiefe):
        """
        Setzt einen Markierungspfeil auf den Vorratsstapel
        :param index: der Index in den Vorratsstapel
        :param tiefe: die anzuzeigende Rekursionstiefe
        """
        h = sum(self.vorratsStapel[i].HoeheGeben() for i in range(index)) + self.vorratsStapel[index].HoeheGeben() // 2
        if tiefe == len(self.pfeile) + 1:
            self.pfeile.append(PfeilSymbol(450 - h, tiefe))
        else:
            self.pfeile[tiefe-1].HoeheSetzen(450 - h)
        self.AufTaktimpulsWarten()

    def PfeilLoeschen(self):
        """
        Löscht den obersten Pfeil.
        """
        alt = self.pfeile.pop()
        alt.Entfernen()
        
class StapelAufgabe:
    """
    Rekursive Lösung eines Rucksackproblems.
    """
    def __init__(self):
        """
        Der Konstruktor erstellt die Paletten und besetzt die Attribute vor
        """
        self.paletten = [
            Palette(80, 9), Palette(90, 8), Palette(50, 5),
            Palette(55, 6), Palette(15, 3), Palette(30, 2), Palette(25, 4)
        ]
        self.optimaleLoesung = None
        self.optimalesGewicht = 0
        self.maximaleHoehe = 150
        self.verwaltung = Stapelverwaltung(self.paletten, self.maximaleHoehe)

    def MaximumsucheStarten(self):
        """
        Sucht die Palettenkonfiguration mit der maximalen Höhe
        """
        self.Schritt(0, 0, 0, [], 1)
        print(f"Das maximal erreichbare Gewicht ist: {self.optimalesGewicht}")

    def Schritt(self, aktIndex, aktHoehe, aktGewicht, loesungsansatz, tiefe):
        """
        Führt den nächsten Schritt der Suche aus
        :param aktIndex: der Index, ab dem die Paletten ausprobiert werden sollen
        :param aktHoehe: die aktuelle Höhe des Stapels
        :param aktGewicht: das aktuelle Gewicht des Stapels
        :param loesungsansatz: Feld mit den bisher auf dem Stapel befindlichen Paletten
        :param tiefe: die Rekursionstiefe
        """
        if aktGewicht > self.optimalesGewicht:
            self.optimalesGewicht = aktGewicht
            self.optimaleLoesung = loesungsansatz[:]
            self.verwaltung.OptimalenStapelSetzen()

        if aktIndex < len(self.paletten):
            for i in range(aktIndex, len(self.paletten)):
                p = self.paletten[i]
                self.verwaltung.PfeilSetzen(i, tiefe)
                if p.HoeheGeben() + aktHoehe <= self.maximaleHoehe:
                    loesungsansatz.append(p)
                    self.verwaltung.AktuellHinzufügen(i)
                    self.Schritt(i + 1, aktHoehe + p.HoeheGeben(), aktGewicht + p.GewichtGeben(), loesungsansatz, tiefe + 1)
                    loesungsansatz.pop()
                    self.verwaltung.AktuellEntfernen()
            self.verwaltung.PfeilLoeschen()

class Rahmen:
    """
    Rahmenklasse des Programms.
    """
    def __init__(self):
        StapelAufgabe().MaximumsucheStarten()

if __name__ == "__main__":
    Rahmen()
